package ru.dbotthepony.mc.otm.client.render

import com.google.gson.JsonDeserializationContext
import com.google.gson.JsonDeserializer
import com.google.gson.JsonElement
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import com.google.gson.JsonSerializationContext
import com.google.gson.JsonSerializer
import com.google.gson.JsonSyntaxException
import com.google.gson.TypeAdapter
import com.google.gson.internal.bind.TypeAdapters
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonWriter
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.resources.ResourceLocation
import ru.dbotthepony.mc.otm.core.set
import java.lang.reflect.Type

fun ResourceLocation.element(
	x: Float,
	y: Float,
	width: Float,
	height: Float,
	textureWidth: Float = 256f,
	textureHeight: Float = 256f
) = SkinElement(this, x, y, width, height, textureWidth, textureHeight)

fun ResourceLocation.pixel(
	x: Float,
	y: Float,
	textureWidth: Float = 256f,
	textureHeight: Float = 256f
) = SkinElement(this, x, y, 1f, 1f, textureWidth, textureHeight)

fun ResourceLocation.hLine(
	x: Float,
	y: Float,
	width: Float,
	textureWidth: Float = 256f,
	textureHeight: Float = 256f
) = SkinElement(this, x, y, width, 1f, textureWidth, textureHeight)

fun ResourceLocation.vLine(
	x: Float,
	y: Float,
	height: Float,
	textureWidth: Float = 256f,
	textureHeight: Float = 256f
) = SkinElement(this, x, y, 1f, height, textureWidth, textureHeight)

@Suppress("unused")
data class SkinElement @JvmOverloads constructor(
	override val texture: ResourceLocation,
	val x: Float,
	val y: Float,
	override val width: Float,
	override val height: Float,
	val imageWidth: Float = 256f,
	val imageHeight: Float = 256f,
	override val winding: UVWindingOrder = UVWindingOrder.NORMAL,
) : AbstractSkinElement() {
	init {
		require(x >= 0f) { "Invalid x $x" }
		require(y >= 0f) { "Invalid y $y" }
		require(width > 0f) { "Invalid width $width" }
		require(height > 0f) { "Invalid height $height" }
		require(imageWidth > 0f) { "Invalid image width $imageWidth" }
		require(imageHeight > 0f) { "Invalid image height $imageHeight" }
	}

	override val u0 = this.x / imageWidth
	override val v0 = this.y / imageHeight
	override val u1 = (this.x + width) / imageWidth
	override val v1 = (this.y + height) / imageHeight

	override val type: SkinElementType
		get() = SkinElementType.SINGLE
}
