package ru.dbotthepony.mc.otm.client.render

import net.minecraft.client.renderer.texture.TextureAtlas
import net.minecraft.client.renderer.texture.TextureAtlasSprite
import net.minecraft.client.renderer.texture.TextureManager
import net.minecraft.client.resources.TextureAtlasHolder
import net.minecraft.resources.ResourceLocation
import net.minecraft.server.packs.resources.ResourceManager
import net.minecraft.util.profiling.ProfilerFiller
import net.minecraftforge.client.event.RegisterClientReloadListenersEvent
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.isClientThread
import ru.dbotthepony.mc.otm.core.WriteOnce
import java.util.stream.Stream
import kotlin.properties.Delegates

var isWidgetAtlasAvailable: Boolean = false
	private set

class WidgetAtlasHolder private constructor(manager: TextureManager) : TextureAtlasHolder(manager, LOCATION, "gui") {
	var changeset = 0
		private set

	var once = false
		private set
	var demandsRebuild = false
		private set

	private var resourceManager by Delegates.notNull<ResourceManager>()
	private var profileManager by Delegates.notNull<ProfilerFiller>()

	override fun prepare(p_118891_: ResourceManager, p_118892_: ProfilerFiller): TextureAtlas.Preparations {
		resourceManager = p_118891_
		profileManager = p_118892_
		changeset++
		return super.prepare(p_118891_, p_118892_)
	}

	override fun apply(p_118894_: TextureAtlas.Preparations, p_118895_: ResourceManager, p_118896_: ProfilerFiller) {
		once = true
		resourceManager = p_118895_
		profileManager = p_118896_
		changeset++
		super.apply(p_118894_, p_118895_, p_118896_)
		AtlasSkinElement.notifyListeners()
	}

	fun rebuildIfRequired(): Boolean {
		if (once && demandsRebuild && isClientThread()) {
			apply(prepare(resourceManager, profileManager), resourceManager, profileManager)
		}

		return demandsRebuild
	}

	public override fun getSprite(p_118902_: ResourceLocation): TextureAtlasSprite {
		if (!once) {
			throw IllegalStateException("Trying to get sprite too early")
		}

		return super.getSprite(p_118902_)
	}

	fun queueRebuild() {
		demandsRebuild = true
	}

	override fun getResourcesToLoad(): Stream<ResourceLocation> {
		demandsRebuild = false
		return AtlasSkinElement.keysStream
	}

	companion object {
		@JvmStatic
		val LOCATION = ResourceLocation(OverdriveThatMatters.MOD_ID, "textures/atlas/gui.png")

		@JvmStatic
		var INSTANCE: WidgetAtlasHolder by WriteOnce()
			private set

		@JvmStatic
		fun register(event: RegisterClientReloadListenersEvent) {
			INSTANCE = WidgetAtlasHolder(minecraft.textureManager)
			event.registerReloadListener(INSTANCE)
			isWidgetAtlasAvailable = true
		}

		@JvmStatic
		fun renderGameHook() {
			if (isWidgetAtlasAvailable) {
				INSTANCE.rebuildIfRequired()
			}
		}
	}
}
