package ru.dbotthepony.mc.otm.menu

import com.google.common.collect.ImmutableList
import com.mojang.datafixers.util.Pair
import it.unimi.dsi.fastutil.objects.Reference2ObjectOpenHashMap
import net.minecraft.resources.ResourceLocation
import net.minecraft.server.level.ServerPlayer
import net.minecraft.world.Container
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.*
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.enchantment.EnchantmentHelper.hasBindingCurse
import net.minecraft.world.level.block.entity.BlockEntity
import net.minecraftforge.network.PacketDistributor
import ru.dbotthepony.mc.otm.capability.matteryPlayer
import ru.dbotthepony.mc.otm.compat.cos.cosmeticArmorSlots
import ru.dbotthepony.mc.otm.compat.curios.isCurioSlot
import ru.dbotthepony.mc.otm.container.ItemFilter
import ru.dbotthepony.mc.otm.container.ItemFilterNetworkSlot
import ru.dbotthepony.mc.otm.menu.widget.AbstractWidget
import ru.dbotthepony.mc.otm.network.FieldSynchronizer
import ru.dbotthepony.mc.otm.network.MenuFieldPacket
import ru.dbotthepony.mc.otm.network.MenuNetworkChannel
import java.util.*

abstract class MatteryMenu @JvmOverloads protected constructor(
	menuType: MenuType<*>?,
	containerId: Int,
	val inventory: Inventory,
	val tile: BlockEntity? = null
) : AbstractContainerMenu(menuType, containerId) {
	val mSynchronizer = FieldSynchronizer()
	val ply: Player get() = inventory.player

	private val _matteryWidgets = ArrayList<AbstractWidget>()
	val matteryWidgets: List<AbstractWidget> = Collections.unmodifiableList(_matteryWidgets)

	private val _playerInventorySlots = ArrayList<MatterySlot>()
	private val _playerInventorySlots2 = ArrayList<MatterySlot>()
	private val _playerHotbarSlots = ArrayList<MatterySlot>()
	private val _playerMainSlots = ArrayList<MatterySlot>()
	private val _playerExoSuitSlots = ArrayList<MatterySlot>()
	private val _playerCombinedInventorySlots = ArrayList<MatterySlot>()

	/**
	 * inventory + exosuit + hotbar (in this order)
	 */
	val playerInventorySlots: List<MatterySlot> = Collections.unmodifiableList(_playerInventorySlots)

	/**
	 * hotbar + inventory + exosuit (in this order)
	 */
	val playerInventorySlots2: List<MatterySlot> = Collections.unmodifiableList(_playerInventorySlots2)

	/**
	 * hotbar only
	 */
	val playerHotbarSlots: List<MatterySlot> = Collections.unmodifiableList(_playerHotbarSlots)

	/**
	 * inventory only
	 */
	val playerMainSlots: List<MatterySlot> = Collections.unmodifiableList(_playerMainSlots)

	/**
	 * exosuit only
	 */
	val playerExoSuitSlots: List<MatterySlot> = Collections.unmodifiableList(_playerExoSuitSlots)

	/**
	 * inventory + exosuit (in this order)
	 */
	val playerCombinedInventorySlots: List<MatterySlot> = Collections.unmodifiableList(_playerCombinedInventorySlots)

	var autoCreateInventoryFrame = true
		private set

	protected val lockedInventorySlots: MutableSet<Int> = HashSet()
	protected open fun isInventorySlotLocked(index: Int): Boolean = lockedInventorySlots.contains(index)

	private val _filterSlots = ArrayList<ItemFilterNetworkSlot>()
	val filterSlots: List<ItemFilterNetworkSlot> = Collections.unmodifiableList(_filterSlots)

	protected var inventorySlotIndexStart = 0
	protected var inventorySlotIndexEnd = 0

	private val playerPacketDistributor = PacketDistributor.PLAYER.with { ply as ServerPlayer }

	fun addFilterSlots(slots: ItemFilter): List<ItemFilterNetworkSlot> {
		val result = ArrayList<ItemFilterNetworkSlot>(slots.size)

		for (i in 0 until slots.size) {
			_filterSlots.add(ItemFilterNetworkSlot(i, _filterSlots.size, slots).also(result::add))
		}

		return result
	}

	fun addFilterSlots(amount: Int): List<ItemFilterNetworkSlot> {
		val result = ArrayList<ItemFilterNetworkSlot>(amount)

		for (i in 0 until amount) {
			_filterSlots.add(ItemFilterNetworkSlot(i, _filterSlots.size, null).also(result::add))
		}

		return result
	}

	fun addWidget(widget: AbstractWidget): Int {
		val indexOf = _matteryWidgets.indexOf(widget)

		if (indexOf == -1) {
			_matteryWidgets.add(widget)
			return _matteryWidgets.size - 1
		}

		return indexOf
	}

	fun getWidget(index: Int): AbstractWidget? {
		if (index !in _matteryWidgets.indices) {
			return null
		}

		return _matteryWidgets[index]
	}

	open inner class InventorySlot(container: Container, index: Int, x: Int = 0, y: Int = 0) : MatterySlot(container, index, x, y) {
		override fun mayPlace(itemStack: ItemStack): Boolean {
			return super.mayPlace(itemStack) && !isInventorySlotLocked(index)
		}

		override fun mayPickup(player: Player): Boolean {
			return super.mayPickup(player) && !isInventorySlotLocked(index)
		}

		override fun isSameInventory(other: Slot): Boolean {
			if (container === inventory || container === ply.matteryPlayer?.exoPackContainer)
				return other.container === inventory || other.container === ply.matteryPlayer?.exoPackContainer

			return super.isSameInventory(other)
		}
	}

	open inner class EquipmentSlot(container: Container, index: Int, val type: net.minecraft.world.entity.EquipmentSlot, x: Int = 0, y: Int = 0) : InventorySlot(container, index, x, y) {
		constructor(type: net.minecraft.world.entity.EquipmentSlot, x: Int = 0, y: Int = 0) : this(
			inventory, 34 + type.ordinal, type, x, y
		)

		override fun mayPlace(itemStack: ItemStack): Boolean {
			return super.mayPlace(itemStack) && itemStack.canEquip(type, inventory.player)
		}

		override fun mayPickup(player: Player): Boolean {
			return super.mayPickup(player) && !item.isEmpty && (player.isCreative || !hasBindingCurse(item))
		}

		override fun getMaxStackSize(): Int {
			return 1
		}

		override fun getNoItemIcon(): Pair<ResourceLocation, ResourceLocation>? {
			return Pair.of(InventoryMenu.BLOCK_ATLAS, TEXTURE_EMPTY_SLOTS[type.index])
		}
	}

	protected fun addInventorySlots(autoFrame: Boolean = !ply.isSpectator) {
		check(_playerInventorySlots.isEmpty()) { "Already created inventory slots" }
		autoCreateInventoryFrame = autoFrame

		for (i in 9 .. 35) {
			val slot = InventorySlot(inventory, i)

			_playerInventorySlots.add(slot)
			_playerInventorySlots2.add(slot)
			_playerMainSlots.add(slot)
			_playerCombinedInventorySlots.add(slot)
			inventory2Storage(slot)
			addSlot(slot)
		}

		val mattery = ply.matteryPlayer

		if (mattery != null && mattery.hasExoPack) {
			for (i in 0 until mattery.exoPackContainer.containerSize) {
				val slot = InventorySlot(mattery.exoPackContainer, i)

				_playerInventorySlots.add(slot)
				_playerInventorySlots2.add(slot)
				_playerExoSuitSlots.add(slot)
				_playerCombinedInventorySlots.add(slot)
				inventory2Storage(slot)
				addSlot(slot)
			}
		}

		for (i in 0..8) {
			val slot = InventorySlot(inventory, i)

			addSlot(slot)
			_playerInventorySlots2.add(i, slot)
			_playerInventorySlots.add(slot)
			_playerHotbarSlots.add(slot)
			inventory2Storage(slot)
		}
	}

	override fun broadcastChanges() {
		for (widget in _matteryWidgets) {
			widget.updateServer()
		}

		val payload = mSynchronizer.collectNetworkPayload()

		if (payload != null) {
			MenuNetworkChannel.send(ply, MenuFieldPacket(payload))
		}

		super.broadcastChanges()

		val consumer = PacketDistributor.PLAYER.with { ply as ServerPlayer }

		for (slot in _filterSlots) {
			val packet = slot.sendChanges()

			if (packet != null) {
				MenuNetworkChannel.send(consumer, packet)
			}
		}
	}

	override fun broadcastFullState() {
		for (widget in _matteryWidgets) {
			widget.updateServer()
		}

		mSynchronizer.invalidate()
		val payload = mSynchronizer.collectNetworkPayload()

		if (payload != null) {
			MenuNetworkChannel.send(ply, MenuFieldPacket(payload))
		}

		super.broadcastFullState()

		val consumer = PacketDistributor.PLAYER.with { ply as ServerPlayer }

		for (slot in _filterSlots) {
			val packet = slot.sendChanges(true)

			if (packet != null) {
				MenuNetworkChannel.send(consumer, packet)
			}
		}
	}

	override fun stillValid(player: Player): Boolean {
		if (tile == null) return true

		if (player.level.getBlockEntity(tile.blockPos) !== tile) {
			return false
		}

		val pos = tile.blockPos

		return player.distanceToSqr(pos.x.toDouble() + 0.5, pos.y.toDouble() + 0.5, pos.z.toDouble() + 0.5) <= 64.0
	}

	abstract val storageSlots: Collection<Slot>
	protected val quickMoveMapping = Reference2ObjectOpenHashMap<Slot, () -> Collection<Slot>?>()

	protected fun storage2Inventory(slot: Slot) {
		quickMoveMapping[slot] = ::playerInventorySlots
	}

	protected fun inventory2Storage(slot: Slot) {
		quickMoveMapping[slot] = ::storageSlots
	}

	protected var autoRegisteredStorageSlots = false

	// This method receive Player interactor and slot_index where Shift + Right click occurred
	// It shall return item stack that got moved
	override fun quickMoveStack(ply: Player, slotIndex: Int): ItemStack {
		if (!slots[slotIndex].hasItem() || !slots[slotIndex].mayPickup(ply)) {
			return ItemStack.EMPTY
		}

		if (!autoRegisteredStorageSlots) {
			autoRegisteredStorageSlots = true

			for (slot in storageSlots) {
				storage2Inventory(slot)
			}
		}

		val slot = slots[slotIndex]
		val target = quickMoveMapping[slot]?.invoke() ?: return ItemStack.EMPTY
		val copy = slots[slotIndex].item.copy()

		if (moveItemStackTo(slots[slotIndex], target))
			return copy
		else
			return ItemStack.EMPTY
	}

	fun quickMoveToInventory(stack: ItemStack, simulate: Boolean): ItemStack {
		if (_playerInventorySlots.isEmpty()) {
			return stack
		}

		return moveItemStackToSlots(stack, _playerInventorySlots, simulate = simulate)
	}

	override fun canTakeItemForPickAll(p_38908_: ItemStack, p_38909_: Slot): Boolean {
		if (p_38909_ is EquipmentSlot)
			return false

		return super.canTakeItemForPickAll(p_38908_, p_38909_) && (p_38909_ !is MatterySlot || p_38909_.canTakeItemForPickAll()) && !p_38909_.isCurioSlot
	}

	override fun moveItemStackTo(
		item: ItemStack,
		initialSlot: Int,
		finalSlot: Int,
		inverse: Boolean
	): Boolean {
		val remainder = moveItemStackToSlots(item, initialSlot, finalSlot, inverse)

		if (remainder.count == item.count) {
			return false
		}

		item.count = remainder.count
		return true
	}

	fun moveItemStackTo(
		item: ItemStack,
		slots: Collection<Slot>
	): Boolean {
		val remainder = moveItemStackToSlots(item, slots)

		if (remainder.count == item.count) {
			return false
		}

		item.count = remainder.count
		return true
	}

	fun moveItemStackTo(
		source: Slot,
		slots: Collection<Slot>
	): Boolean {
		val remainder = moveItemStackToSlots(source.item, slots)

		if (remainder.count == source.item.count) {
			return false
		}

		if (remainder.isEmpty) {
			source.set(ItemStack.EMPTY)
		} else {
			source.item.count = remainder.count
			source.setChanged()
		}

		return true
	}

	fun moveItemStackToSlots(item: ItemStack, slots: Collection<Slot>, simulate: Boolean = false): ItemStack {
		val copy = item.copy()

		// first pass - stack with existing slots
		if (copy.isStackable) {
			for (slot in slots) {
				val limit = slot.getMaxStackSize(copy)

				if (limit > slot.item.count && ItemStack.isSameItemSameTags(slot.item, copy) && slot.mayPlace(item)) {
					val newCount = (slot.item.count + copy.count).coerceAtMost(limit)
					val diff = newCount - slot.item.count
					copy.count -= diff

					if (!simulate) {
						slot.item.count += diff
						slot.setChanged()
					}

					if (copy.isEmpty) {
						return copy
					}
				}
			}
		}

		// second pass - drop stack into first free slot
		for (slot in slots) {
			val limit = slot.getMaxStackSize(copy)

			if (!slot.hasItem() && slot.mayPlace(item)) {
				val newCount = copy.count.coerceAtMost(limit)

				if (!simulate) {
					slot.set(copy.copy().also { it.count = newCount })
					slot.setChanged()
				}

				copy.count -= newCount

				if (copy.isEmpty) {
					return copy
				}
			}
		}

		return copy
	}

	fun moveItemStackToSlots(item: ItemStack, initialSlot: Int, finalSlot: Int, inverse: Boolean = false, simulate: Boolean = false): ItemStack {
		if (initialSlot > finalSlot) {
			return item
		}

		require(initialSlot >= 0) { "Invalid initial slot $initialSlot" }
		require(finalSlot < slots.size) { "Final slot $finalSlot is bigger than total size of array of ${slots.size}" }

		val slots = ArrayList<Slot>(finalSlot - initialSlot + 1)

		for (i in (if (inverse) finalSlot downTo initialSlot else initialSlot .. finalSlot)) {
			slots.add(slots[i])
		}

		return moveItemStackToSlots(item, slots, simulate)
	}

	public override fun addDataSlots(p_38885_: ContainerData) {
		super.addDataSlots(p_38885_)
	}

	fun makeArmorSlots(): List<kotlin.Pair<EquipmentSlot, Slot?>> {
		val cosmetic = ply.cosmeticArmorSlots

		return ImmutableList.of(
			EquipmentSlot(net.minecraft.world.entity.EquipmentSlot.HEAD) to cosmetic?.get(net.minecraft.world.entity.EquipmentSlot.HEAD),
			EquipmentSlot(net.minecraft.world.entity.EquipmentSlot.CHEST) to cosmetic?.get(net.minecraft.world.entity.EquipmentSlot.CHEST),
			EquipmentSlot(net.minecraft.world.entity.EquipmentSlot.LEGS) to cosmetic?.get(net.minecraft.world.entity.EquipmentSlot.LEGS),
			EquipmentSlot(net.minecraft.world.entity.EquipmentSlot.FEET) to cosmetic?.get(net.minecraft.world.entity.EquipmentSlot.FEET),
		)
	}

	companion object {
		val TEXTURE_EMPTY_SLOTS: List<ResourceLocation> = ImmutableList.of(
			InventoryMenu.EMPTY_ARMOR_SLOT_BOOTS,
			InventoryMenu.EMPTY_ARMOR_SLOT_LEGGINGS,
			InventoryMenu.EMPTY_ARMOR_SLOT_CHESTPLATE,
			InventoryMenu.EMPTY_ARMOR_SLOT_HELMET)
	}
}
